#include <arduinoFFT.h>
#include <LiquidCrystal_I2C.h>
#include <math.h>

extern "C" {
  #include "adc.h"
}

#define SAMPLES 128
#define SAMPLING_FREQUENCY 2048
#define SAMPLING_PERIOD round(1000000 * (1.0 / SAMPLING_FREQUENCY))
#define LCD_ADDR 0x27
#define LCD_COLS 16
#define LCD_ROWS 2
#define NUM_OF_NOTES 63
#define A4_FREQ 440
#define A4_INDEX 48
#define THRESHOLD 800.0
#define FFT_ERROR 3.0
#define REF_RNG 1.0

ArduinoFFT<float> FFT = ArduinoFFT<float>();
LiquidCrystal_I2C LCD = LiquidCrystal_I2C(LCD_ADDR, LCD_COLS, LCD_ROWS);

unsigned long microseconds;

float vReal[SAMPLES];
float vImag[SAMPLES];

const char* notes[] = {
	"A", "A#", "B", "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#"
};

float freq[NUM_OF_NOTES];

void gen_freq() {
	for (int i = 0; i < NUM_OF_NOTES; i++) {
		freq[i] = A4_FREQ * pow(2, (float)(i - A4_INDEX) / 12.0);
	}
}

float get_fft_error(float f) {
	if (f >= freq[0] && f < 200.0) {
		return 2.0;
	} else if (f >= 200.0 && f < 300.0) {
		return 4.0;
	} else if (f >= 300.0 && f < 400.0) {
		return 6.0;
	} else if (f >= 300.0 && f < 400.0) {
		return 8.0;
	} else {
		return 10.0;
	}
}

void create_tuner_chars() {
	byte lcd_char[8] = {
		0b000000, 0b000000, 0b000000, 0b000000,
		0b000000, 0b000000, 0b000000, 0b111111
	};

	byte addr = 0x0;

	for (int i = 6; i >= 0; i--) {
		lcd_char[i] = 0b111111;
		LCD.createChar(addr, lcd_char);
		addr++;
	}
}

void turn_on_color(char color) {
	PORTD = 0;

	if (color == 'r') {
		PORTD |= (1 << PD4);
	} else if (color == 'g') {
		PORTD |= (1 << PD3);
	} else if (color == 'b') {
		PORTD |= (1 << PD2);
	}
}

void setup() {
	DDRD |= (1 << PD2) | (1 << PD3) | (1 << PD4);
	adc_init();
	Serial.begin(9600);
	LCD.init();
	LCD.backlight();
	LCD.setCursor(0, 0);
	LCD.print("Guitar Tuner v1");
	gen_freq();
	create_tuner_chars();
	delay(1000);
}

void loop() {
	for (int i = 0; i < SAMPLES; i++) {
		microseconds = micros();
		
		vReal[i] = static_cast<float>(get_sample(PC0)) - 512.0;
		vImag[i] = 0;
		
		while (micros() < (microseconds + SAMPLING_PERIOD)) {}
	}

	FFT.windowing(vReal, SAMPLES, FFT_WIN_TYP_HANN, FFT_FORWARD);
	FFT.compute(vReal, vImag, SAMPLES, FFT_FORWARD);
	FFT.complexToMagnitude(vReal, vImag, SAMPLES);

	float peak;
	float magnitude;

	FFT.majorPeak(vReal, SAMPLES, SAMPLING_FREQUENCY, &peak, &magnitude);
	peak -= get_fft_error(peak);

	if (magnitude > THRESHOLD) {
		int closest = round(12 * log(peak / A4_FREQ) / log(2.0)) + A4_INDEX;

		if (closest < 0) {
			closest = 0;
		} else if (closest > NUM_OF_NOTES) {
			closest = NUM_OF_NOTES;
		}


		float ref = freq[closest];

		if (peak >= ref - REF_RNG && peak <= ref + REF_RNG) {
			turn_on_color('g');
			LCD.clear();
			LCD.setCursor(7, 0);
			LCD.print(notes[closest % 12]);
			LCD.setCursor(7, 1);
			LCD.print("OK");
		} else {
			LCD.clear();
			LCD.setCursor(7, 0);
			LCD.print(notes[closest % 12]);
			LCD.setCursor(7, 1);
			LCD.print("<>");
			
			turn_on_color('r');

			if (peak < ref) {
				float lower;
				closest - 1 >= 0 ? lower = freq[closest - 1] : lower = 0;

				if (lower) {
					float step = (ref - lower) / 14.0;
					uint16_t seg_num = (uint16_t)((ref - peak) / step);
					if (seg_num < 7) seg_num++;

					for (uint16_t seg = 0; seg < seg_num; seg++) {
						LCD.setCursor(6 - seg, 1);
						LCD.write(6 - seg);
					}
				}
			}
			else {
				float higher;
				closest + 1 <= 63 ? higher = freq[closest + 1] : higher = 0;

				if (higher) {
					float step = (higher - ref) / 14.0;
					uint16_t seg_num = (uint16_t)((peak - ref) / step);
					if (seg_num < 7) seg_num++;

					Serial.println(step);
					Serial.println(seg_num);
					Serial.println();

					for (uint16_t seg = 0; seg < seg_num; seg++) {
						LCD.setCursor(9 + seg, 1);
						LCD.write(6 - seg);
					}
				}
			}
		}

		delay(500);
	} else {
		turn_on_color('b');
		delay (500);
		LCD.clear();
		LCD.setCursor(0, 0);
		LCD.print("Play some notes");
	}
}
